"""
Spectate
========

.. automodule:: spectate.spectate
    :members:
    :ignore-module-all:

.. automodule:: spectate.mvc
    :members:
"""

from .spectate import *
from ._version import __version__
